import logging
l = logging.getLogger('ana.d')

def D(uuid, child_cls, state):
    l.debug("Deserializing Storable with uuid %s", uuid)

    if uuid is None and state is None:
        raise ANAError("A blank UUID and a blank state were passed into the deserialization routine. This is most likely caused by a failure to return anything from object._ana_getstate().")

    if uuid is not None:
        self = get_dl().uuid_cache.get(uuid, None)
        if self is not None:
            l.debug("... returning cached")
            return self

    self = super(Storable, child_cls).__new__(child_cls) #pylint:disable=bad-super-call
    dl = get_dl()
    if uuid is not None:
        dl.uuid_cache[uuid] = self

    if uuid is not None and state is None:
        l.debug("... loading state")
        state = get_dl().load_state(uuid)

    if uuid is not None:
        self._stored = dl.uuid
        l.debug("... returning newly cached")
    else:
        self._stored = None
        l.debug("... returning non-UUID storable")

    self._ana_setstate(state)
    self._ana_uuid = uuid

    if not hasattr(self, '_ana_uuid'):
        raise ANAError("Storable somehow got through without an _ana_uuid attr")
    return self

D.__safe_for_unpickling__ = True

from .storable import Storable
from .errors import ANAError
from . import get_dl



import uuid as uuid_module

import logging
l = logging.getLogger('ana.storable')

def _all_slots(cls):
    return sum((o.__slots__ for o in cls.mro() if hasattr(o, '__slots__') and not o is Storable),[])

class Storable(object):
    __slots__ = [ '_ana_uuid', '_stored', '__weakref__' ]

    def make_uuid(self, uuid=None):
        '''
        If the storable has no UUID, this function creates one. The UUID is then
        returned.
        '''
        u = getattr(self, '_ana_uuid', None)
        if u is None:
            u = str(uuid_module.uuid4()) if uuid is None else uuid
            l.debug("Caching UUID %s", u)
            get_dl().uuid_cache[u] = self
            setattr(self, '_ana_uuid', u)
        return u

    @property
    def ana_uuid(self):
        return self.make_uuid()

    def ana_store(self):
        '''
        Assigns a UUID to the storable and stores the actual data.
        '''
        u = self.make_uuid()
        dl = get_dl()
        if not getattr(self, '_stored', None) == dl.uuid:
            state = self._ana_getstate()
            setattr(self, '_stored', dl.uuid)
            dl.store_state(u, state)
        return u

    @classmethod
    def ana_load(cls, uuid):
        return D(uuid, cls, get_dl().load_state(uuid))

    @staticmethod
    def _any_to_literal(o, known_set, objects):
        if o is None:
            return None
        elif type(o) in (int, str, float, bool):
            return o
        elif isinstance(o, dict):
            return {
                Storable._any_to_literal(k, known_set, objects):Storable._any_to_literal(v, known_set, objects) for k,v in o.items()
            }
        elif isinstance(o, list) or isinstance(o, tuple) or isinstance(o, set):
            return [ Storable._any_to_literal(e, known_set, objects) for e in o ]
        elif isinstance(o, Storable):
            return o._self_to_literal(known_set, objects)
        else:
            if hasattr(o, '__getstate__'):
                state = o.__getstate__()
            elif hasattr(o, '__dict__'):
                state = o.__dict__
            else:
                state = { k: getattr(o, k) for k in _all_slots(o.__class__)  }

            return {
                'class': o.__class__.__name__,
                'object': Storable._any_to_literal(state, known_set, objects)
            }

    def _self_to_literal(self, known_set, objects):
        uuid = self.make_uuid()

        if uuid not in known_set:
            known_set.add(uuid)
            o = self._ana_getliteral()
            objects[uuid] = {
                #'module': getattr(self, '__module__', '__unknown__'),
                'class': self.__class__.__name__,
                'object': self._any_to_literal(o, known_set, objects)
            }

        return { 'ana_uuid': uuid }

    def to_literal(self, known_set=None, objects=None):
        if known_set is None:
            known_set = set()

        objects = { } if objects is None else objects
        return { 'objects': objects, 'value': self._self_to_literal(known_set, objects) }

    #
    # ANA API
    #

    @classmethod
    def _all_slots(cls):
        #pylint:disable=no-member
        return _all_slots(cls)

    def _ana_getstate(self):
        if hasattr(self, '__dict__'):
            return self.__dict__
        else:
            return { k: getattr(self, k) for k in self._all_slots() if hasattr(self, k) }

    def _ana_setstate(self, s):
        for k,v in s.items():
            setattr(self, k, v)

    def _ana_getliteral(self):
        return self._ana_getstate()

    #
    # Pickle API
    #

    def __reduce__(self):
        u = getattr(self, '_ana_uuid', None)
        if u is None:
            return (D, (None, self.__class__, self._ana_getstate()))
        elif get_dl()._store_type == 'simple':
            return (D, (u, self.__class__, self._ana_getstate()))
        else:
            self.ana_store()
            return (D, (u, self.__class__, None))


from . import get_dl
from .d import D



import os
import uuid as uuid_module
import weakref

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import pymongo
    import bson
except ImportError:
    # mongo dependency is optional
    pymongo = None

import logging
l = logging.getLogger("ana.datalayer")

class DataLayer(object):
    '''
    The DataLayer handles storing and retrieving UUID-identified objects
    to/from a central store.
    '''

    def __init__(self):
        self.uuid_cache = weakref.WeakValueDictionary()
        self._store_type = None
        self.uuid = uuid_module.uuid4()

    def store_state(self, uuid, s):
        raise NotImplementedError()

    def load_state(self, uuid):
        raise NotImplementedError()

class SimpleDataLayer(DataLayer):
    def __init__(self):
        DataLayer.__init__(self)
        self._store_type = 'simple'

    def store_state(self, uuid, s):
        return

    def load_state(self, uuid):
        raise ANAError("SimpleDataLayer does not support state loading.")

class DirDataLayer(DataLayer):
    def __init__(self, pickle_dir):
        DataLayer.__init__(self)
        self._store_type = 'pickle'
        self._dir = pickle_dir

        if not os.path.exists(self._dir):
            l.warning("Directory '%s' doesn't exit. Creating.", self._dir)
            os.makedirs(self._dir)

    def store_state(self, uuid, s):
        with open(os.path.join(self._dir, str(uuid)+'.p'), 'wb') as f:
            pickle.dump(s, f, protocol=pickle.HIGHEST_PROTOCOL)

    def load_state(self, uuid):
        with open(os.path.join(self._dir, str(uuid)+'.p'), 'rb') as f:
            return pickle.load(f)

class MongoDataLayer(DataLayer):
    def __init__(self, mongo_args, mongo_db='ana', mongo_collection='storage'):
        DataLayer.__init__(self)
        if pymongo is None:
            raise ImportError("pymongo necessary for ANA mongo backend")

        l.debug("Pickling into mongo.")

        self._store_type = 'mongo'
        self._mongo = pymongo.MongoClient(*mongo_args)[mongo_db][mongo_collection]

    def store_state(self, uuid, s):
        # TODO: investigate whether check/insert is faster than
        # upsert (because of latency) and also deal with the race
        # condition here
        if self._mongo.find({'_id': uuid}).limit(1).count(with_limit_and_skip=True) == 0:
            p = pickle.dumps(s, protocol=pickle.HIGHEST_PROTOCOL)
            self._mongo.insert_one({'_id': uuid, 'pickled': bson.binary.Binary(p)})

    def load_state(self, uuid):
        p = self._mongo.find_one({'_id': uuid})['pickled']
        return pickle.loads(p)

class DictDataLayer(DataLayer):
    def __init__(self, the_dict=None):
        DataLayer.__init__(self)
        self._store_type = 'dict'
        self._state_store = { } if the_dict is None else the_dict

    def store_state(self, uuid, s):
        p = pickle.dumps(s, protocol=pickle.HIGHEST_PROTOCOL)
        self._state_store[uuid] = p

    def load_state(self, uuid):
        p = self._state_store[uuid]
        return pickle.loads(p)

from .errors import ANAError



#pylint:disable=wildcard-import

import logging
logging.getLogger("ana").addHandler(logging.NullHandler())

from .datalayer import *

dl = SimpleDataLayer()
def get_dl():
    return dl
def set_dl(new_dl):
    global dl
    dl = new_dl

class M(object):
    '''This is a marker that's used internally by ANA.'''
    __slots__ = [ ]

from .storable import *
